/* Copyright 2010 ESRI
* 
* All rights reserved under the copyright laws of the United States
* and applicable international laws, treaties, and conventions.
* 
* You may freely redistribute and use this sample code, with or
* without modification, provided you include the original copyright
* notice and use restrictions.
* 
* See the use restrictions at http://help.arcgis.com/en/sdk/10.0/usageRestrictions.htm
* 
*/
package com.esri.arcgis.sample.simplerouting;

import com.esri.adf.web.data.WebContext;
import com.esri.adf.web.faces.event.TaskEvent;
import com.esri.adf.web.data.GeocodeFunctionality;
import com.esri.adf.web.data.results.WebResults;
import com.esri.adf.web.data.AddressCandidate;
import com.esri.adf.web.data.GISResource;
import com.esri.adf.web.data.tasks.TaskParamDescriptor;
import com.esri.adf.web.data.tasks.TaskParamDescriptorModel;
import com.esri.adf.web.ags.data.AGSNAFunctionality;
import com.esri.adf.web.ags.data.AGSLocalExtensionsResource;
import com.esri.adf.web.ags.data.AGSExtensionsResource;
import com.esri.adf.web.ags.data.AGSMapResource;
import com.esri.adf.web.data.geometry.WebPoint;
import com.esri.arcgisws.EsriNetworkForwardStarBacktrack;
import com.esri.adf.web.ags.data.AGSTraversalResult;
import com.esri.adf.web.ags.data.AGSDirection;
import com.esri.adf.web.data.tasks.TaskActionDescriptor;
import com.esri.adf.web.data.tasks.TaskUtils;
import com.esri.arcgisws.NAServerLocations;
import com.esri.arcgisws.NAServerPropertySets;
import com.esri.arcgisws.NAServerRouteParams;
import com.esri.arcgisws.PropertySet;
import com.esri.arcgisws.PropertySetProperty;
import com.esri.adf.web.data.WebContextObserver;
import com.esri.adf.web.data.tasks.TaskInfoBean;
import com.esri.adf.web.data.tasks.TaskInfo;

import java.util.ArrayList;
import java.util.Map;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;

public final class SimpleRoutingTask implements WebContextObserver, TaskInfoBean {
  private TaskInfo taskInfo = new SimpleRoutingTaskInfo();
  private WebContext wContext = null;
  private GeocodeFunctionality geoFunc = null;
  private AGSNAFunctionality naFunc = null;
  private AGSMapResource naResource = null;
  private static final Logger logger = Logger.getLogger(SimpleRoutingTask.class.getName());
  private String taskName = "Sample Routing Task";
  private String errorMessage = null;
  private boolean hideErrorMessage = true;

  private void initTask(GISResource resource) {
    logger.fine("SampleRoutingTask.initTask");
    taskInfo.getTaskDescriptor().setDisabled(false);
    if (resource == null) {
      logger.log(Level.SEVERE, "GISResource is null");
      taskInfo.getTaskDescriptor().setDisabled(true);
      return;
    }
    geoFunc = (GeocodeFunctionality) resource
        .getFunctionality(GeocodeFunctionality.FUNCTIONALITY_NAME);
    if (geoFunc == null) {
      logger.log(Level.SEVERE, "GeocodeFunctionality not found in GISResource: " + resource);
      taskInfo.getTaskDescriptor().setDisabled(true);
      return;
    }

    Map resources = wContext.getResources();
    Object item = null;
    for (Iterator iter = resources.entrySet().iterator(); iter.hasNext();) {
      item = ((java.util.Map.Entry) iter.next()).getValue();
      if (item instanceof AGSLocalExtensionsResource) {
        naResource = (AGSLocalExtensionsResource) item;
        naFunc = (AGSNAFunctionality) naResource
            .getFunctionality(AGSNAFunctionality.FUNCTIONALITY_NAME);
      } else if (item instanceof AGSExtensionsResource) {
        naResource = (AGSExtensionsResource) item;
        naFunc = (AGSNAFunctionality) naResource
            .getFunctionality(AGSNAFunctionality.FUNCTIONALITY_NAME);
      }
    }
    if (naFunc == null) {
      logger.log(Level.SEVERE, "AGSNAFunctionality not found in GISResource: " + naResource);
      taskInfo.getTaskDescriptor().setDisabled(true);
      return;
    }

    initParams();
  }

  private void initParams() {
    logger.fine("SampleRoutingTask.initParams");
    String[] fields = geoFunc.getAddressFields();
    if (fields == null)
      return;
    ((SimpleRoutingTaskInfo) taskInfo).setParamDescriptors(fields);
  }

  /**
   * update is called every time sever context is modified. Helps to enable and
   * disable different buttons on the tool.
   *
   * @see com.esri.adf.web.data.WebContextObserver#update(com.esri.adf.web.data.WebContext,
   *      java.lang.Object)
   */
  public void update(WebContext context, Object arg) {
    toggleDisablity();
  }

  /**
   * return the name of the task
   */
  public String getTaskName() {
    return taskName;
  }

  /**
   * sets the name of the task
   */
  public void setTaskName(String taskName) {
    this.taskName = taskName;
  }

  /**
   * returns the error message
   */
  public String getErrorMessage() {
    return errorMessage;
  }

  private void augmentLastStopWithAddress(String address) {
    NAServerRouteParams routeParams = naFunc.getRouteParams();
    PropertySet[] stops = ((NAServerPropertySets) routeParams.getStops()).getPropertySets();
    PropertySetProperty[] lastStop = stops[stops.length - 1].getPropertyArray();
    for (int i = 0; i < lastStop.length; i++) {
      if (lastStop[i].getKey().equals("Name")) {
        lastStop[i].setValue(address);
      }
    }
    NAServerPropertySets naServerPropSets = new NAServerPropertySets();
    naServerPropSets.setPropertySets(stops);
    routeParams.setStops(naServerPropSets);
  }

  /**
   * computes the route between the two geographic locations defined by the
   * start and stop addresses
   */
  public void computeRoute(TaskEvent event) {
    if (geoFunc != null && naFunc != null) {
      String[] fields = geoFunc.getAddressFields();
      TaskParamDescriptorModel[] params = taskInfo.getParamDescriptors();
      RoutingTaskParamDescriptor[] prms = new RoutingTaskParamDescriptor[fields.length];
      boolean empty = true;
      for (int i = 0; i < fields.length; i++) {
        prms[i] = (RoutingTaskParamDescriptor) params[i];
        empty = empty
            && ((prms[i].getParamValue(null) == null) || (prms[i].getParamValue(null).toString()
                .length() == 0));

      }
      if (!empty) {
        WebContext wc = event.getWebContext();
        AddressCandidate start = locate(wc, prms);
        String startAddress = "";
        for (int i = 0; i < prms.length; i++) {
          startAddress = startAddress + prms[i].fieldValue + " ";
        }
        if (start == null) {
          hideErrorMessage = false;
          errorMessage = "Starting Address Invalid!";
        }
        for (int i = 0; i < fields.length; i++) {
          prms[i] = (RoutingTaskParamDescriptor) params[i + fields.length];
        }
        AddressCandidate stop = locate(wc, prms);
        String stopAddress = "";
        for (int i = 0; i < prms.length; i++) {
          stopAddress = stopAddress + prms[i].fieldValue + " ";
        }
        if (stop == null) {
          hideErrorMessage = false;
          errorMessage = "Destination Address Invalid!";
        }

        if ((start != null) && (stop != null)) {
          hideErrorMessage = true;
          WebPoint startPoint = (WebPoint) start.getLocation();
          naFunc.addStop(startPoint);
          augmentLastStopWithAddress(startAddress);

          WebPoint stopPoint = (WebPoint) stop.getLocation();
          naFunc.addStop(stopPoint);
          augmentLastStopWithAddress(stopAddress);

          solve(event.getWebContext().getWebResults());
        }

        wContext.refresh();
      }
    }
  }

  private AddressCandidate locate(WebContext wc, RoutingTaskParamDescriptor[] prms) {
    for (int i = 0; i < prms.length; i++) {
      geoFunc.setAddressFieldValue(prms[i].fieldName, (String) prms[i].fieldValue);
    }
    AddressCandidate location = geoFunc.geocodeAddress();
    return location;
  }

  /**
   * clears the route and the input fields
   */
  public void clearRoute(TaskEvent event) {
    if (geoFunc != null || naFunc != null) {
      naFunc.clearStops();
      naFunc.clearRoute();
      initParams();
      wContext.refresh();
    }
  }

  private void solve(WebResults webResults) {
    naFunc.getRouteParams().setIgnoreInvalidLocations(true);
    naFunc.getRouteParams().setRestrictUTurns(
        EsriNetworkForwardStarBacktrack.esriNFSBAllowBacktrack);
    AGSTraversalResult results[] = naFunc.solve();
    if (results != null && results.length == 1) {
      wContext.refresh();
      AGSTraversalResult result = results[0];
      AGSDirection summary = result.getSummary();
      ArrayList directions = (ArrayList) result.getDirections();
      webResults.addResultsWithActionArray(summary.getDirectionText(), directions,
          "getDirectionText", "getDirectionString", null);
    }
  }

  /**
   * returns web context
   */
  public WebContext getWebContext() {
    return wContext;
  }

  /**
   * sets geocode resource. This method is called by the JSF framework when
   * instantiating SimpleRoutingTask managed bean.
   *
   * @param resource
   *          an instance of AGSLocalGeocodeResource
   */
  public void setGeocodeResource(GISResource resource) {
    if (wContext == null)
      taskInfo.getTaskDescriptor().setDisabled(true);
    if (!java.beans.Beans.isDesignTime()) {
      this.wContext = resource.getWebContext();
      this.wContext.addObserver(this);
      initTask(resource);
      toggleDisablity();
    }
  }

  /**
   * returns the TaskInfo object describing SimpleRoutingTask's input fields and
   * action buttons
   */
  public TaskInfo getTaskInfo() {
    if (wContext == null)
      taskInfo.getTaskDescriptor().setDisabled(true);
    return this.taskInfo;
  }

  /**
   * sets the TaskInfo object
   */
  public void setTaskInfo(TaskInfo info) {
    this.taskInfo = info;
  }

  private void toggleDisablity() {
    TaskActionDescriptor clearRoute = (TaskActionDescriptor) TaskUtils.getActionDescriptor(
        "clearRoute", taskInfo);
    TaskActionDescriptor computeRoute = (TaskActionDescriptor) TaskUtils.getActionDescriptor(
        "computeRoute", taskInfo);
    TaskParamDescriptor errorMessageLabel = (TaskParamDescriptor) TaskUtils.getParamDescriptor(
        "errorMessage", taskInfo);
    NAServerLocations locations = null;
    int stopCount = -1;
    if (naFunc == null) {
      return;
    }
    if ((locations = naFunc.getRouteParams().getStops()) != null) {
      stopCount = ((NAServerPropertySets) locations).getPropertySets().length;
    }
    if (stopCount > 1) {
      clearRoute.setDisabled(false);
      computeRoute.setDisabled(true);
    } else {
      clearRoute.setDisabled(true);
      computeRoute.setDisabled(false);
    }
    errorMessageLabel.setHidden(hideErrorMessage);
  }
}
